from unittest import TestCase
from typing import Tuple
from click.testing import CliRunner
from aac.execute.command_line import cli, initialize_cli


from aac.plugins.print_defs.print_aac_definitions_impl import print_defs


class TestPrintAaCDefinitions(TestCase):
    def test_print_defs(self):

        result1 = print_defs(False)
        self.assertTrue(result1.is_success(), msg=f"Expected success but failed with message: {result1.is_success()}")
        self.assertGreater(len(result1.get_messages_as_string()), 0, msg="Expected at least one message but received none.")

        result2 = print_defs(True)
        self.assertTrue(result2.is_success(), msg=f"Expected success but failed with message: {result2.is_success()}")
        self.assertGreater(len(result1.get_messages_as_string()), 0, msg="Expected at least one message but received none.")

    def run_print_defs_cli_command_with_args(self, args: list[str]) -> Tuple[int, str]:
        """Utility function to invoke the CLI command with the given arguments."""
        initialize_cli()
        runner = CliRunner()
        result = runner.invoke(cli, ["print-defs"] + args)
        exit_code = result.exit_code
        std_out = str(result.stdout)
        output_message = std_out.strip().replace("\x1b[0m", "")
        return exit_code, output_message

    def test_cli_print_defs_core_only(self):
        args = ["--core-only"]

        exit_code, output_message = self.run_print_defs_cli_command_with_args(args)
        self.assertEqual(0, exit_code)  # asserts the command ran successfully
        self.assertGreater(len(output_message), 0)  # asserts the command produced output

        self.assertNotIn("name: Exclusive Fields", output_message)
        self.assertNotIn("name: No Extends for Final", output_message)
        self.assertIn("name: AacType", output_message)
        self.assertIn("name: Modifier", output_message)

    def test_cli_print_defs(self):
        args = []

        exit_code, output_message = self.run_print_defs_cli_command_with_args(args)
        self.assertEqual(0, exit_code)  # asserts the command ran successfully
        self.assertGreater(len(output_message), 0)  # asserts the command produced output

        self.assertIn("name: Exclusive Fields", output_message)
        self.assertIn("name: No Extends for Final", output_message)
        self.assertIn("name: AacType", output_message)
        self.assertIn("name: Modifier", output_message)



from io import StringIO
from unittest import TestCase
from unittest.mock import patch
from typing import Tuple
from click.testing import CliRunner
from os import path, mkdir, listdir
from shutil import copy
from tempfile import TemporaryDirectory
from traceback import print_exception

from aac.execute.command_line import cli, initialize_cli


class TestGenPlugin(TestCase):

    def test_gen_plugin(self):
        # I'm going to rely on the CLI testing for this one, but will leave there here in case we need it later
        pass

    def run_gen_plugin_cli_command_with_args(self, args: list[str]) -> Tuple[int, str]:
        """Utility function to invoke the CLI command with the given arguments."""
        initialize_cli()
        runner = CliRunner()
        result = runner.invoke(cli, ["gen-plugin"] + args)
        exit_code = result.exit_code
        std_out = str(result.stdout)
        output_message = std_out.strip().replace("\x1b[0m", "")
        return exit_code, output_message

    def test_cli_gen_plugin(self):
        # first we need a project to work in, so generate a temporary one
        with TemporaryDirectory() as temp_dir:
            aac_file_path = path.join(path.dirname(__file__), "my_project.aac")
            temp_aac_file_path = path.join(temp_dir, "my_project.aac")
            copy(aac_file_path, temp_aac_file_path)

            proj_args = [temp_aac_file_path, "--output", temp_dir, "--no-prompt"]

            exit_code, output_message = self.run_gen_project_cli_command_with_args(proj_args)

            self.assertEqual(0, exit_code, f"Expected success but failed with message: {output_message}")  # asserts the command ran successfully

            # now create an AaC plugin file in the project src directory
            package_src_path = path.join(temp_dir, "src", "happy")
            mkdir(package_src_path)
            package_tests_path = path.join(temp_dir, "tests", "test_happy")
            mkdir(package_tests_path)
            package_doc_path = path.join(temp_dir, "docs")
            plugin_file_path = path.join(package_src_path, "my_plugin.aac")

            aac_plugin_path = path.join(path.dirname(__file__), "my_plugin.aac")
            copy(aac_plugin_path, plugin_file_path)

            plugin_args = [plugin_file_path, "--code-output", path.join(temp_dir, "src"), "--test-output", path.join(temp_dir, "tests"), "--doc-output", package_doc_path,"--no-prompt"]

            exit_code, output_message = self.run_gen_plugin_cli_command_with_args(plugin_args)
            self.assertEqual(0, exit_code)  # asserts the command ran successfully
            self.assertIn("All AaC constraint checks were successful", output_message)  # asserts the command ran check successfully

            # make sure the files were created correctly
            self.assertTrue(path.exists(path.join(package_src_path, "__init__.py")))
            self.assertTrue(path.exists(path.join(package_src_path, "my_plugin_impl.py")))
            self.assertTrue(path.exists(path.join(package_tests_path, "test_my_plugin.py")))
            self.assertTrue(path.exists(path.join(package_tests_path, "my_plugin_command_test.feature")))
            self.assertTrue(path.exists(path.join(package_tests_path, "my_plugin_context_test.feature")))
            self.assertTrue(path.exists(path.join(package_tests_path, "my_plugin_schema_test.feature")))
            self.assertTrue(path.exists(path.join(package_tests_path, "my_plugin_primitive_test.feature")))
            self.assertTrue(path.exists(path.join(package_doc_path, "my_plugin.md")))

            # Ensure feature files were generated with a Given, When, and Then.
            files = listdir(package_tests_path)
            feature_files = list(filter(lambda x: ".feature" in x, files))
            for feature_file in feature_files:
                f = open(path.join(package_tests_path, feature_file), "r")
                file_content = f.read()
                self.assertIn("Given", file_content)
                self.assertIn("When", file_content)
                self.assertIn("Then", file_content)
                f.close()

            # Ensure the feature file steps were assigned to the correct keyword.
            file = open(path.join(package_tests_path, "my_plugin_command_test.feature"), "r")
            file_content = file.read()
            self.assertIn("Given An AaC file containing schemas with no extra fields.", file_content)
            self.assertIn("When The AaC check command is run on the schema.", file_content)
            self.assertIn("Then The check commands provides the output 'All AaC constraint checks were successful.'", file_content)
            file.close()


    def test_cli_gen_plugin_overwrite(self):
        # first we need a project to work in, so generate a temporary one
        with TemporaryDirectory() as temp_dir:
            # now create an AaC plugin file in the project src directory
            aac_file_path = path.join(path.dirname(__file__), "my_project.aac")
            temp_aac_file_path = path.join(temp_dir, "my_project.aac")
            copy(aac_file_path, temp_aac_file_path)

            proj_args = [temp_aac_file_path, "--output", temp_dir, "--no-prompt"]
            exit_code, output_message = self.run_gen_project_cli_command_with_args(proj_args)

            package_src_path = path.join(temp_dir, "src", "happy")
            mkdir(package_src_path)
            package_tests_path = path.join(temp_dir, "tests", "test_happy")
            mkdir(package_tests_path)
            package_doc_path = path.join(temp_dir, "docs")
            plugin_file_path = path.join(package_src_path, "my_plugin.aac")

            aac_plugin_path = path.join(path.dirname(__file__), "my_plugin.aac")
            copy(aac_plugin_path, plugin_file_path)
            plugin_args = [plugin_file_path, "--code-output", path.join(temp_dir, "src"), "--test-output", path.join(temp_dir, "tests"), "--doc-output", package_doc_path, "--no-prompt"]
            exit_code, output_message = self.run_gen_plugin_cli_command_with_args(plugin_args)
            file = open(path.join(package_src_path, "my_plugin_impl.py"), "r")
            file_read = file.read()
            self.assertIn("test_command_one(aac_plugin_file: str)", file_read)
            self.assertNotIn("test_command_1_evaluation(aac_plugin_file: str)", file_read)
            file.close()
            # Check for documentation generation
            file = open(path.join(package_doc_path, "my_plugin.md"), "r")
            file_read = file.read()
            self.assertIn("test-command-one", file_read)
            self.assertNotIn("test-command-1-evaluation", file_read)
            file.close()
            # Confirm no generated .aac_backup files
            self.assertFalse(path.exists(path.join(package_src_path, "my_plugin_impl.py.aac_backup")))

            aac_plugin_path = path.join(path.dirname(__file__), "my_plugin_eval.aac")
            copy(aac_plugin_path, plugin_file_path)
            plugin_args = [plugin_file_path, "--code-output", path.join(temp_dir, "src"), "--test-output", path.join(temp_dir, "tests"), "--doc-output", package_doc_path, "--no-prompt",  "--force-overwrite"]
            exit_code, output_message = self.run_gen_plugin_cli_command_with_args(plugin_args)
            file = open(path.join(package_src_path, "my_plugin_impl.py"), "r")
            file_read = file.read()
            self.assertIn("test_command_1_evaluation(aac_plugin_file: str)", file_read)
            self.assertNotIn("test_command_one(aac_plugin_file: str)", file_read)
            file.close()
            # Check for documentation generation
            file = open(path.join(package_doc_path, "my_plugin.md"), "r") # Note that the name stays as my_plugin because it comes from the name found within my_plugin_eval.aac
            file_read = file.read()
            self.assertIn("test-command-1-evaluation", file_read)
            self.assertNotIn("test-command-one", file_read)
            file.close()
            # Check for generated .aac_backup files
            self.assertTrue(path.exists(path.join(package_src_path, "my_plugin_impl.py.aac_backup")))

    def test_cli_gen_plugin_evaluate(self):
        # first we need a project to work in, so generate a temporary one
        with TemporaryDirectory() as temp_dir:
            # now create an AaC plugin file in the project src directory
            aac_file_path = path.join(path.dirname(__file__), "my_project.aac")
            temp_aac_file_path = path.join(temp_dir, "my_project.aac")
            copy(aac_file_path, temp_aac_file_path)

            proj_args = [temp_aac_file_path, "--output", temp_dir, "--no-prompt"]
            exit_code, output_message = self.run_gen_project_cli_command_with_args(proj_args)

            package_src_path = path.join(temp_dir, "src", "happy")
            mkdir(package_src_path)
            package_tests_path = path.join(temp_dir, "tests", "test_happy")
            mkdir(package_tests_path)
            package_doc_path = path.join(temp_dir, "docs")
            plugin_file_path = path.join(package_src_path, "my_plugin.aac")

            aac_plugin_path = path.join(path.dirname(__file__), "my_plugin.aac")
            copy(aac_plugin_path, plugin_file_path)
            plugin_args = [plugin_file_path, "--code-output", path.join(temp_dir, "src"), "--test-output", path.join(temp_dir, "tests"), "--doc-output", package_doc_path, "--no-prompt"]
            exit_code, output_message = self.run_gen_plugin_cli_command_with_args(plugin_args)
            file = open(path.join(package_src_path, "my_plugin_impl.py"), "r")
            file_read = file.read()
            self.assertIn("test_command_one(aac_plugin_file: str)", file_read)
            self.assertNotIn("test_command_one_evaluation(aac_plugin_file: str)", file_read)
            file.close()
            # Confirm no generated .aac_evaluate files
            self.assertFalse(path.exists(path.join(package_src_path, "my_plugin_impl.py.aac_evaluate")))

            aac_plugin_path = path.join(path.dirname(__file__), "my_plugin_eval.aac")
            copy(aac_plugin_path, plugin_file_path)
            plugin_args = [plugin_file_path, "--code-output", path.join(temp_dir, "src"), "--test-output", path.join(temp_dir, "tests"), "--doc-output", package_doc_path, "--no-prompt", "--evaluate"]
            exit_code, output_message = self.run_gen_plugin_cli_command_with_args(plugin_args)
            file = open(path.join(package_src_path, "my_plugin_impl.py"), "r")
            file_read = file.read()
            self.assertNotIn("test_command_1_evaluation(aac_plugin_file: str)", file_read)
            self.assertIn("test_command_one(aac_plugin_file: str)", file_read)
            file.close()
            # Check for documentation generation
            file = open(path.join(package_doc_path, "my_plugin.md"), "r") # Note that the name stays as my_plugin because it comes from the name found within my_plugin_eval.aac
            file_read = file.read()
            self.assertIn("test-command-one", file_read)
            self.assertNotIn("test-command-1-evaluation", file_read)
            file.close()
            # Check .aac_evaluate file has been created
            self.assertTrue(path.exists(path.join(package_src_path, "my_plugin_impl.py.aac_evaluate")))

    def test_cli_gen_plugin_with_incomplete_dirs(self):
        # first we need a project to work in, so generate a temporary one
        with TemporaryDirectory() as temp_dir:
            aac_file_path = path.join(path.dirname(__file__), "my_project.aac")
            temp_aac_file_path = path.join(temp_dir, "my_project.aac")
            copy(aac_file_path, temp_aac_file_path)

            proj_args = [temp_aac_file_path, "--output", temp_dir, "--no-prompt"]

            exit_code, output_message = self.run_gen_project_cli_command_with_args(proj_args)

            self.assertEqual(0, exit_code, f"Expected success but failed with message: {output_message}")  # asserts the command ran successfully

            # now create an AaC plugin file in the project src directory
            package_src_path = path.join(temp_dir, "src", "happy")
            mkdir(package_src_path)
            plugin_file_path = path.join(temp_dir, "my_plugin.aac")

            aac_plugin_path = path.join(path.dirname(__file__), "my_plugin.aac")
            copy(aac_plugin_path, plugin_file_path)

            plugin_args = [plugin_file_path, "--code-output", path.join(temp_dir, "src"), "--no-prompt"]

            exit_code, output_message = self.run_gen_plugin_cli_command_with_args(plugin_args)
            self.assertEqual(0, exit_code)  # asserts the command ran successfully
            self.assertIn("All AaC constraint checks were successful", output_message)  # asserts the command ran check successfully

            self.assertTrue(path.exists(path.join(package_src_path, "my_plugin_impl.py")))

            package_tests_path = path.join(temp_dir, "test_happy")  # making sure that tests get output to the correct location
            self.assertTrue(path.exists(path.join(package_src_path, "__init__.py")))
            self.assertTrue(path.exists(path.join(package_tests_path, "my_plugin_command_test.feature")))
            self.assertTrue(path.exists(path.join(package_tests_path, "my_plugin_command_test_two.feature")))
            self.assertTrue(path.exists(path.join(package_tests_path, "my_plugin_command_test_three.feature")))

    def test_cli_gen_plugin_multiple_commands(self):
        with TemporaryDirectory() as temp_dir:
            aac_file_path = path.join(path.dirname(__file__), "my_project.aac")
            temp_aac_file_path = path.join(temp_dir, "my_project.aac")
            copy(aac_file_path, temp_aac_file_path)

            proj_args = [temp_aac_file_path, "--output", temp_dir, "--no-prompt"]

            exit_code, output_message = self.run_gen_project_cli_command_with_args(proj_args)

            self.assertEqual(0, exit_code, f"Expected success but failed with message: {output_message}")  # asserts the command ran successfully

            # now create an AaC plugin file in the project src directory
            package_src_path = path.join(temp_dir, "src", "happy")
            mkdir(package_src_path)
            package_tests_path = path.join(temp_dir, "tests", "test_happy")
            mkdir(package_tests_path)
            plugin_file_path = path.join(package_src_path, "my_plugin.aac")

            aac_plugin_path = path.join(path.dirname(__file__), "my_plugin.aac")
            copy(aac_plugin_path, plugin_file_path)

            plugin_args = [plugin_file_path, "--code-output", path.join(temp_dir, "src"), "--test-output", path.join(temp_dir, "tests"), "--no-prompt"]

            exit_code, output_message = self.run_gen_plugin_cli_command_with_args(plugin_args)
            self.assertEqual(0, exit_code)  # asserts the command ran successfully
            self.assertIn("All AaC constraint checks were successful", output_message)  # asserts the command ran check successfully

            self.assertTrue(path.exists(path.join(package_src_path, "my_plugin_impl.py")))
            with open(path.join(package_src_path, "my_plugin_impl.py")) as impl_file:
                impl_file_read = impl_file.read()
                self.assertIn("def before_test_command_one_check", impl_file_read)
                self.assertIn("def test_command_one", impl_file_read)
                self.assertIn("def after_test_command_one_generate", impl_file_read)

                self.assertIn("def before_test_command_two_check", impl_file_read)
                self.assertIn("def test_command_two", impl_file_read)
                self.assertIn("def after_test_command_two_generate", impl_file_read)

                self.assertIn("def before_test_command_three_check", impl_file_read)
                self.assertIn("def test_command_three", impl_file_read)
                self.assertIn("def after_test_command_three_generate", impl_file_read)

            self.assertTrue(path.exists(path.join(package_src_path, "__init__.py")))
            with open(path.join(package_src_path, "__init__.py")) as init_file:
                init_file_read = init_file.read()
                self.assertIn("def run_test_command_one", init_file_read)

                self.assertIn("def run_test_command_two", init_file_read)

                self.assertIn("def run_test_command_three", init_file_read)
            self.assertTrue(path.exists(path.join(package_tests_path, "my_plugin_command_test.feature")))
            self.assertTrue(path.exists(path.join(package_tests_path, "my_plugin_command_test_two.feature")))
            self.assertTrue(path.exists(path.join(package_tests_path, "my_plugin_command_test_three.feature")))

    # The following decorators are to stop std_out and std_err from clogging up the terminal with System errors.
    @patch('sys.stdout', new_callable=StringIO)  # Suppress stdout
    @patch('sys.stderr', new_callable=StringIO)  # Suppress stderr
    def test_cli_gen_plugin_failure(self, mock_stderr, mock_stdout):
        with TemporaryDirectory() as temp_dir:
            aac_file_path = path.join(path.dirname(__file__), "my_plugin_bad.aac")
            temp_aac_file_path = path.join(temp_dir, "my_plugin_bad.aac")
            copy(aac_file_path, temp_aac_file_path)

            proj_args = [temp_aac_file_path, "--code-output", temp_dir, "--no-prompt"]

            exit_code, output_message = self.run_gen_plugin_cli_command_with_args(proj_args)
            self.assertNotEqual(0, exit_code)
            self.assertIn("Definition is missing field 'name'", output_message)

    def test_gen_project(self):
        # I'm going to rely on the CLI testing for this one, but will leave there here in case we need it later
        pass

    def run_gen_project_cli_command_with_args(self, args: list[str]) -> Tuple[int, str]:
        """Utility function to invoke the CLI command with the given arguments."""
        initialize_cli()
        runner = CliRunner()
        result = runner.invoke(cli, ["gen-project"] + args)
        if result.exception:
            exc_type, exc_value, exc_traceback = result.exc_info
            print_exception(exc_type, exc_value, exc_traceback)
        exit_code = result.exit_code
        std_out = str(result.stdout)
        output_message = std_out.strip().replace("\x1b[0m", "")
        return exit_code, output_message

    def test_cli_gen_project(self):
        with TemporaryDirectory() as temp_dir:
            aac_file_path = path.join(path.dirname(__file__), "my_project.aac")
            temp_aac_file_path = path.join(temp_dir, "my_project.aac")
            copy(aac_file_path, temp_aac_file_path)

            args = [temp_aac_file_path, "--output", temp_dir, "--no-prompt"]

            exit_code, output_message = self.run_gen_project_cli_command_with_args(args)

            self.assertEqual(0, exit_code, f"Expected success but failed with message: {output_message}")  # asserts the command ran successfully
            self.assertIn("All AaC constraint checks were successful", output_message)  # asserts the command ran check successfully

            # make sure the files and directories were created
            self.assertTrue(path.exists(path.join(temp_dir, "pyproject.toml")))
            self.assertTrue(path.exists(path.join(temp_dir, "tox.ini")))
            self.assertTrue(path.exists(path.join(temp_dir, "README.md")))
            self.assertTrue(path.exists(path.join(temp_dir, "src")))
            self.assertTrue(path.exists(path.join(temp_dir, "tests")))
            self.assertTrue(path.exists(path.join(temp_dir, "docs")))

    # The following decorators are to stop std_out and std_err from clogging up the terminal with System errors.
    @patch('sys.stdout', new_callable=StringIO)  # Suppress stdout
    @patch('sys.stderr', new_callable=StringIO)  # Suppress stderr
    def test_cli_gen_project_failure(self, mock_stderr, mock_stdout):
        with TemporaryDirectory() as temp_dir:
            aac_file_path = path.join(path.dirname(__file__), "my_project_bad.aac")
            temp_aac_file_path = path.join(temp_dir, "my_project_bad.aac")
            copy(aac_file_path, temp_aac_file_path)

            proj_args = [temp_aac_file_path, "--output", temp_dir, "--no-prompt"]

            exit_code, output_message = self.run_gen_project_cli_command_with_args(proj_args)
            self.assertNotEqual(0, exit_code)
            self.assertIn("Definition is missing field 'name'", output_message)

    # There is a special case where documentation needs to be generated for optional arguments
    # This tests for that special case and the expected hard-coded header of "Optional Arguments"
    # The other tests in this file account for the nominal cases
    def test_cli_gen_plugin_optional_args_documentation(self):
        # first we need a project to work in, so generate a temporary one
        with TemporaryDirectory() as temp_dir:
            # now create an AaC plugin file in the project src directory
            aac_file_path = path.join(path.dirname(__file__), "gen_plugin_optional_args.aac")
            temp_aac_file_path = path.join(temp_dir, "gen_plugin_optional_args.aac")
            copy(aac_file_path, temp_aac_file_path)

            proj_args = [temp_aac_file_path, "--output", temp_dir, "--no-prompt"]
            exit_code, output_message = self.run_gen_project_cli_command_with_args(proj_args)

            package_src_path = path.join(temp_dir, "src", "happy")
            mkdir(package_src_path)
            package_doc_path = path.join(temp_dir, "docs")
            plugin_file_path = path.join(package_src_path, "gen_plugin_optional_args.aac")

            aac_plugin_path = path.join(path.dirname(__file__), "gen_plugin_optional_args.aac")
            copy(aac_plugin_path, plugin_file_path)
            plugin_args = [plugin_file_path, "--code-output", path.join(temp_dir, "src"), "--test-output", path.join(temp_dir, "tests"), "--doc-output", package_doc_path, "--no-prompt"]
            exit_code, output_message = self.run_gen_plugin_cli_command_with_args(plugin_args)
            file = open(path.join(package_doc_path, "gen_plugin.md"), "r")
            file_read = file.read()
            self.assertIn("Optional Arguments", file_read) # Optional header to test for
            self.assertIn("--doc-output", file_read) # Optional argument from gen_plugin_optional_args.aac
            file.close()

    # There is a special case where documentation needs to be generated but the plugin has no arguments
    # This tests for that special case and the expected hard-coded language of "There are no arguments for the..."
    # The other tests in this file account for the nominal cases
    def test_cli_gen_plugin_no_args_documentation(self):
        # first we need a project to work in, so generate a temporary one
        with TemporaryDirectory() as temp_dir:
            # now create an AaC plugin file in the project src directory
            aac_file_path = path.join(path.dirname(__file__), "version_no_args.aac")
            temp_aac_file_path = path.join(temp_dir, "version_no_args.aac")
            copy(aac_file_path, temp_aac_file_path)

            proj_args = [temp_aac_file_path, "--output", temp_dir, "--no-prompt"]
            exit_code, output_message = self.run_gen_project_cli_command_with_args(proj_args)

            package_src_path = path.join(temp_dir, "src", "happy")
            mkdir(package_src_path)
            package_doc_path = path.join(temp_dir, "docs")
            plugin_file_path = path.join(package_src_path, "version_no_args.aac")

            aac_plugin_path = path.join(path.dirname(__file__), "version_no_args.aac")
            copy(aac_plugin_path, plugin_file_path)
            plugin_args = [plugin_file_path, "--code-output", path.join(temp_dir, "src"), "--test-output", path.join(temp_dir, "tests"), "--doc-output", package_doc_path, "--no-prompt"]
            exit_code, output_message = self.run_gen_plugin_cli_command_with_args(plugin_args)
            file = open(path.join(package_doc_path, "version.md"), "r")
            file_read = file.read()
            self.assertIn("There are no arguments for the", file_read)
            file.close()



"""The AaC AaC primitive constraints plugin implementation module."""
# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists.

from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    MessageLevel,
)
from aac.in_out.files.aac_file import AaCFile
from aac.context.source_location import SourceLocation
from aac.context.language_context import LanguageContext
from aac.context.language_error import LanguageError

from os import linesep
from datetime import datetime


plugin_name = "AaC Primitive Constraints"


def check_bool(
    value: str, type_declaration: str, source: AaCFile, location: SourceLocation
) -> ExecutionResult:
    """Business logic for the Check bool constraint."""

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []

    if not isinstance(value, bool):
        status = ExecutionStatus.CONSTRAINT_FAILURE
        error_msg = ExecutionMessage(
            message=f"{value} is not a valid bool value.",
            level=MessageLevel.ERROR,
            source=(source.uri if source else "No file to reference"),
            location=location,
        )
        messages.append(error_msg)

    return ExecutionResult(plugin_name, "Check bool", status, messages)


def check_date(
    value: str, type_declaration: str, source: AaCFile, location: SourceLocation
) -> ExecutionResult:
    """Business logic for the Check date constraint."""

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []

    try:
        datetime.fromisoformat(str(value))
    except ValueError as error:
        status = ExecutionStatus.CONSTRAINT_FAILURE
        error_msg = ExecutionMessage(
            message=f"Failed to parse the date time string '{value}' with error: '{error}'",
            level=MessageLevel.ERROR,
            source=(source.uri if source else "No file to reference"),
            location=location,
        )
        messages.append(error_msg)

    return ExecutionResult(plugin_name, "Check date", status, messages)


def check_directory(
    value: str, type_declaration: str, source: AaCFile, location: SourceLocation
) -> ExecutionResult:
    """Business logic for the Check directory constraint."""

    # I think this should work fine
    return check_file(value, type_declaration, source, location)


def check_file(  # noqa: C901
    value: str, type_declaration: str, source: AaCFile, location: SourceLocation
) -> ExecutionResult:
    """Business logic for the Check file constraint."""

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []

    if not isinstance(value, str):
        status = ExecutionStatus.CONSTRAINT_FAILURE
        error_msg = ExecutionMessage(
            message=f"{value} is not a valid file path. Value should be string but received {type(value)}",
            level=MessageLevel.ERROR,
            source=(source.uri if source else "No file to reference"),
            location=location,
        )
        messages.append(error_msg)
        return ExecutionResult(plugin_name, "Check file", status, messages)

    # I tried to do this with a regex but got frustrated and gave up.  For some reason, the complex regex the AI generated for me would just run forever when matching.
    # So here's a brute force approach.
    valid_separators = ["/", "\\"]
    item = ""
    path_entries = []
    valid_path_characters = (
        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_."
    )
    for character in value:
        if character in valid_separators:
            if item != "":
                path_entries.append(item)
                item = ""
        else:
            item += character
    path_entries.append(item)

    first = True
    for path_entry in path_entries:
        # check for relative path entries (. and ..)
        if path_entry == "." or path_entry == "..":
            continue
        if first and path_entry.endswith(":"):
            # this is a windows drive letter, so make sure it's valid
            first = False
            if len(path_entry) != 2:
                status = ExecutionStatus.CONSTRAINT_FAILURE
                error_msg = ExecutionMessage(
                    message=f"The value {value} has an invalid drive specification.",
                    level=MessageLevel.ERROR,
                    source=(source.uri if source else "No file to reference"),
                    location=location,
                )
                messages.append(error_msg)
                return ExecutionResult(plugin_name, "Check file", status, messages)
            if (
                path_entry[0]
                not in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
            ):
                status = ExecutionStatus.CONSTRAINT_FAILURE
                error_msg = ExecutionMessage(
                    message=f"The value {value} has an invalid drive specification.",
                    level=MessageLevel.ERROR,
                    source=(source.uri if source else "No file to reference"),
                    location=location,
                )
                messages.append(error_msg)
                return ExecutionResult(plugin_name, "Check file", status, messages)
            continue
        else:
            # this needs to be a valid directory or file name
            for character in path_entry:
                if character not in valid_path_characters:
                    status = ExecutionStatus.CONSTRAINT_FAILURE
                    error_msg = ExecutionMessage(
                        message=f"The value {value} is not a valid file path.  Invalid character '{character}'.",
                        level=MessageLevel.ERROR,
                        source=(source.uri if source else "No file to reference"),
                        location=location,
                    )
                    messages.append(error_msg)
                    return ExecutionResult(plugin_name, "Check file", status, messages)

    # no bad content found, so return success
    return ExecutionResult(plugin_name, "Check file", status, messages)


def check_string(
    value: str, type_declaration: str, source: AaCFile, location: SourceLocation
) -> ExecutionResult:
    """Business logic for the Check string constraint."""

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []

    if not isinstance(value, str):
        status = ExecutionStatus.CONSTRAINT_FAILURE
        error_msg = ExecutionMessage(
            message=f"{value} is not a valid string value.  Type declaration = {type_declaration}",
            level=MessageLevel.ERROR,
            source=(source.uri if source else "No file to reference"),
            location=location,
        )
        messages.append(error_msg)

    return ExecutionResult(plugin_name, "Check string", status, messages)


def check_int(
    value: str, type_declaration: str, source: AaCFile, location: SourceLocation
) -> ExecutionResult:
    """Business logic for the Check int constraint."""

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []

    is_invalid = False
    try:
        if isinstance(value, int):
            return ExecutionResult(plugin_name, "Check int", status, messages)
        type_casted_int = int(value)
        # assert that the conversion didn't alter the contents, like in the case of float -> int.
        is_invalid = str(type_casted_int) != str(value)
    except Exception as error:
        is_invalid = True
        status = ExecutionStatus.CONSTRAINT_FAILURE
        error_msg = ExecutionMessage(
            message=f"AaC int primitive constraint failed for value {value} with error:{linesep}{error}",
            level=MessageLevel.ERROR,
            source=(source.uri if source else "No file to reference"),
            location=location,
        )
        messages.append(error_msg)

    if is_invalid:
        status = ExecutionStatus.CONSTRAINT_FAILURE
        error_msg = ExecutionMessage(
            message=f"{value} is not a valid value for the primitive type int",
            level=MessageLevel.ERROR,
            source=(source.uri if source else "No file to reference"),
            location=location,
        )
        messages.append(error_msg)

    return ExecutionResult(plugin_name, "Check int", status, messages)


def check_number(
    value: str, type_declaration: str, source: AaCFile, location: SourceLocation
) -> ExecutionResult:
    """Business logic for the Check number constraint."""

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []

    is_invalid = False
    try:
        if isinstance(value, int):
            return ExecutionResult(plugin_name, "Check number", status, messages)
        if isinstance(value, float):
            return ExecutionResult(plugin_name, "Check number", status, messages)
        type_casted_float = float(value)
        # assert that the conversion didn't alter the contents, other than adding a '.0' for a whole number value.
        is_invalid = str(type_casted_float) != str(value) and str(
            type_casted_float
        ) != str(value + ".0")
    except Exception as error:
        is_invalid = True
        status = ExecutionStatus.CONSTRAINT_FAILURE
        error_msg = ExecutionMessage(
            message=f"AaC number primitive constraint failed for value {value} with error:{linesep}{error}",
            level=MessageLevel.ERROR,
            source=(source.uri if source else "No file to reference"),
            location=location,
        )
        messages.append(error_msg)

    if is_invalid:
        status = ExecutionStatus.CONSTRAINT_FAILURE
        error_msg = ExecutionMessage(
            message=f"{value} is not a valid value for the primitive type number",
            level=MessageLevel.ERROR,
            source=(source.uri if source else "No file to reference"),
            location=location,
        )
        messages.append(error_msg)

    return ExecutionResult(plugin_name, "Check number", status, messages)


def check_dataref(
    value: str, type_declaration: str, source: AaCFile, location: SourceLocation
) -> ExecutionResult:
    """Business logic for the Check dataref constraint."""

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []

    clean_value = str(value).strip()
    if clean_value.endswith("[]"):
        clean_value = value[:-2].strip()

    # get the dataref target from the parenthesis in the type_declaration
    dataref_target = type_declaration[
        type_declaration.find("(") + 1: type_declaration.find(")")
    ]
    if not dataref_target:
        status = ExecutionStatus.CONSTRAINT_FAILURE
        error_msg = ExecutionMessage(
            message=f"Dataref constraint '{type_declaration}' failed for value '{clean_value}' with error: No dataref target provided.",
            level=MessageLevel.ERROR,
            source=(source.uri if source else "No file to reference"),
            location=location,
        )
        messages.append(error_msg)

    context: LanguageContext = LanguageContext()
    found_values = context.get_values_by_field_chain(dataref_target)
    if clean_value not in found_values:
        status = ExecutionStatus.CONSTRAINT_FAILURE
        error_msg = ExecutionMessage(
            message=f"Dataref constraint failed for value '{clean_value}': '{clean_value}' not found in '{dataref_target}' values '{found_values}'",
            level=MessageLevel.ERROR,
            source=(source.uri if source else "No file to reference"),
            location=location,
        )
        messages.append(error_msg)

    return ExecutionResult(plugin_name, "Check dataref", status, messages)


def check_typeref(
    value: str, type_declaration: str, source: AaCFile, location: SourceLocation
) -> ExecutionResult:
    """Business logic for the Check typeref constraint."""
    if not type_declaration.startswith("typeref"):
        raise Exception(
            f"Invalid typeref constraint: '{value}' for declaration '{type_declaration}' for source '{source.uri}' at location '{location}'"
        )

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []

    clean_value = str(value).strip()
    if clean_value.endswith("[]"):
        clean_value = value[:-2].strip()

    if "(" in clean_value:
        clean_value = clean_value[: clean_value.find("(")].strip()

    context: LanguageContext = LanguageContext()

    qualified_type_name = type_declaration[
        type_declaration.find("(") + 1: type_declaration.find(")")
    ]
    parts = qualified_type_name.split(".")
    package_name = ".".join(parts[:-1])
    type_name = parts[-1]

    try:
        definitions_of_type = context.get_definitions_of_type(package_name, type_name)
    except LanguageError as e:
        raise LanguageError(e.message, source.uri if source else "No file to reference")
    value_defining_type = context.get_definitions_by_name(clean_value)

    if len(value_defining_type) != 1:
        status = ExecutionStatus.CONSTRAINT_FAILURE
        error_msg = ExecutionMessage(
            message=f"Typeref constraint {type_declaration} failed for value {clean_value} with error: No unique defining schema found for value.",
            level=MessageLevel.ERROR,
            source=(source.uri if source else "No file to reference"),
            location=location,
        )
        messages.append(error_msg)
    else:
        value_root_key = value_defining_type[0].get_root_key()
        try:
            root_type_definition = context.get_defining_schema_for_root(value_root_key)
        except LanguageError as e:
            raise LanguageError(e.message, value_defining_type[0].source.uri)
        if root_type_definition not in definitions_of_type:
            status = ExecutionStatus.CONSTRAINT_FAILURE
            error_msg = ExecutionMessage(
                message=f"Typeref constraint {type_declaration} failed for value {clean_value} with error: Value {clean_value} is not a valid {qualified_type_name}.",
                level=MessageLevel.ERROR,
                source=(source.uri if source else "No file to reference"),
                location=location,
            )
            messages.append(error_msg)

    return ExecutionResult(plugin_name, "Check typeref", status, messages)



from unittest import TestCase
from aac.context.lexeme import Lexeme
from aac.context.source_location import SourceLocation


class TestLexeme(TestCase):
    def test_lexeme(self):
        loc = SourceLocation(1, 2, 3, 4)
        src = "my_source.aac"
        value = "my_value"

        lex_one = Lexeme(loc, src, value)
        lex_two = Lexeme(loc, src, value)

        self.assertEqual(lex_one, lex_two)

        string_value = str(lex_one)
        self.assertIsInstance(string_value, str)
        self.assertGreater(len(string_value), 0)

    def test_lexeme_fail(self):
        loc = SourceLocation(1, 2, 3, 4)
        src = "my_source.aac"
        value = "my_value"

        lex_one = Lexeme(loc, src, value)
        lex_two = Lexeme(loc, value, src)

        self.assertNotEqual(lex_one, lex_two)



"""The AaC Check AaC plugin test module."""
from unittest import TestCase
from typing import Tuple
from click.testing import CliRunner
from aac.execute.command_line import cli, initialize_cli
import os
import shutil
import tempfile


class TestCheckAaC(TestCase):
    """Class to test the Check AaC plugin."""
    # FIX ME: Several tests in this class should be updated when the LanguageError and ParserError exceptions thrown by the check_aac_impl.py call to
    # FIX ME: parse_and_load are handled properly.

    # some default test
    def test_check(self):
        """Incomplete test method to directly test 'check'."""
        # Let's just test via CLI for now
        pass

    # The test helper method in this class which invokes the CliRunner for all of the other tests
    def run_check_cli_command_with_args(self, args: list[str]) -> Tuple[int, str]:
        """Utility function to invoke the CLI command with the given arguments."""
        initialize_cli()
        runner = CliRunner()
        result = runner.invoke(cli, ["check"] + args)
        exit_code = result.exit_code
        std_out = str(result.stdout)
        output_message = std_out.strip().replace("\x1b[0m", "")
        return exit_code, output_message

    # Happy path test without verbose flag
    # Test input passes all tests in parsing and check_aac_impl.py
    def test_cli_check(self):
        """Test the CLI command for the check plugin."""
        with tempfile.TemporaryDirectory() as temp_dir:
            aac_file_path = os.path.join(os.path.dirname(__file__), "good.aac")
            temp_aac_file_path = os.path.join(temp_dir, "my_plugin.aac")
            shutil.copy(aac_file_path, temp_aac_file_path)

            check_args = [temp_aac_file_path]

            exit_code, output_message = self.run_check_cli_command_with_args(check_args)
            self.assertEqual(0, exit_code, f"Expected success but failed with message: {output_message}")  # asserts the command ran successfully
            self.assertNotIn("My plugin was successful.", output_message)  # only appears when --verbose is passed in.
            self.assertIn("my_plugin.aac", output_message)

    # Happy path test with verbose flag
    # Test input passes all tests in parsing and check_aac_impl.py
    def test_cli_check_verbose(self):
        """Test the CLI command for the check plugin with the verbose flag on."""
        with tempfile.TemporaryDirectory() as temp_dir:
            aac_file_path = os.path.join(os.path.dirname(__file__), "good.aac")
            temp_aac_file_path = os.path.join(temp_dir, "my_plugin.aac")
            shutil.copy(aac_file_path, temp_aac_file_path)

            check_args = [temp_aac_file_path, "--verbose"]

            exit_code, output_message = self.run_check_cli_command_with_args(check_args)
            self.assertEqual(0, exit_code, f"Expected success but failed with message: {output_message}")  # asserts the command ran successfully
            self.assertIn("My plugin was successful.", output_message)  # only appears when --verbose is passed in.
            self.assertIn("my_plugin.aac", output_message)

    # Test input triggers a LanguageError in check_aac_impl.py ~line 171
    # Value of 'parent_specs' was expected to be list, but was '<class 'str'>'
    # This test should be looking for a LanguageError or message content from a LanguageError but the exceptions from context.parse_and_load(aac_file)
    # in check_aac_imply.py are not being caught and handled gracefully.
    def test_cli_check_bad_data(self):
        """Test check_aac_impl.py and trigger a LanguageError."""
        with tempfile.TemporaryDirectory() as temp_dir:
            aac_file_path = os.path.join(os.path.dirname(__file__), "bad.aac")
            temp_aac_file_path = os.path.join(temp_dir, "my_plugin.aac")
            shutil.copy(aac_file_path, temp_aac_file_path)

            check_args = [temp_aac_file_path]

            exit_code, output_message = self.run_check_cli_command_with_args(check_args)
            self.assertNotEqual(0, exit_code)
            self.assertIn("was expected to be list, but was", output_message)
            self.assertIn("my_plugin.aac", output_message)

    # Test input improper YAML
    # output msg: The AaC file '/tmp/{random-temp-name}/my_plugin.aac' could not be parsed.
    # This test should be looking for a ParserError or message content from a ParserError but the exceptions from context.parse_and_load(aac_file)
    # in check_aac_imply.py are not being caught and handled gracefully.
    def test_parse_Error(self):
        """Test check_aac_impl.py with improper YAML."""
        with tempfile.TemporaryDirectory() as temp_dir:
            aac_file_path = os.path.join(os.path.dirname(__file__), "parse_error.aac")
            temp_aac_file_path = os.path.join(temp_dir, "my_plugin.aac")
            shutil.copy(aac_file_path, temp_aac_file_path)

            check_args = [temp_aac_file_path]

            exit_code, output_message = self.run_check_cli_command_with_args(check_args)
            # Assert for return code 3 (fail)
            self.assertEqual(3, exit_code, f"Expected to fail but ran successfully with message: {output_message}")
            self.assertNotIn("My plugin was successful.", output_message)  # only appears when --verbose is passed in.
            self.assertIn("my_plugin.aac", output_message)
            self.assertIn("invalid YAML", output_message)

    # Test input missing required 'when' field primitive
    # output msg: Missing required field when.
    def test_unexpected_primitive(self):
        """Test check_aac_impl.py with a missing required field."""
        with tempfile.TemporaryDirectory() as temp_dir:
            aac_file_path = os.path.join(os.path.dirname(__file__), "missing_required_field.aac")
            temp_aac_file_path = os.path.join(temp_dir, "my_plugin.aac")
            shutil.copy(aac_file_path, temp_aac_file_path)

            check_args = [temp_aac_file_path]

            exit_code, output_message = self.run_check_cli_command_with_args(check_args)
            # Assert for return code 6 (missing required field)
            self.assertEqual(6, exit_code, f"Expected to fail but ran successfully with message: {output_message}")
            self.assertNotIn("My plugin was successful.", output_message)  # only appears when --verbose is passed in.
            self.assertIn("Missing required field when.", output_message)


    # Test input with an unknown primitive "stranger" which triggers a ParserError
    # output msg: Found undefined field name 'stranger' when expecting ['name', 'tags', 'requirements', 'given', 'when', 'then', 'examples'] as defined in Scenario
    # This test should be looking for a ParserError or message content from a ParserError but the exceptions from context.parse_and_load(aac_file)
    # in check_aac_imply.py are not being caught and handled gracefully.
    def test_unknown_primitive(self):
        """Test check_aac_impl.py with an unknown primitive resulting in a ParserError."""
        with tempfile.TemporaryDirectory() as temp_dir:
            aac_file_path = os.path.join(os.path.dirname(__file__), "unknown_field.aac")
            temp_aac_file_path = os.path.join(temp_dir, "my_plugin.aac")
            shutil.copy(aac_file_path, temp_aac_file_path)

            check_args = [temp_aac_file_path]

            exit_code, output_message = self.run_check_cli_command_with_args(check_args)
            # Assert for return code 6 (unexpected primitive)
            self.assertEqual(6, exit_code, f"Expected to fail but ran successfully with message: {output_message}")
            self.assertNotIn("My plugin was successful.", output_message)  # only appears when --verbose is passed in.
            self.assertIn("my_plugin.aac", output_message)

    # Test input with an an invalid value
    # output msg: Invalid value for field 'name'.  Expected type 'str', but found '<class 'list'>'
    def test_invalid_value(self):
        """Test check_aac_impl.py with an invalid value."""
        with tempfile.TemporaryDirectory() as temp_dir:
            aac_file_path = os.path.join(os.path.dirname(__file__), "invalid_value.aac")
            temp_aac_file_path = os.path.join(temp_dir, "my_plugin.aac")
            shutil.copy(aac_file_path, temp_aac_file_path)

            check_args = [temp_aac_file_path]

            exit_code, output_message = self.run_check_cli_command_with_args(check_args)
            # Assert for return code 6 (invalid value)
            self.assertEqual(6, exit_code, f"Expected to fail but ran successfully with message: {output_message}")
            self.assertNotIn("My plugin was successful.", output_message)  # only appears when --verbose is passed in.
            self.assertIn("Invalid value for field", output_message)

    # Test input with an an undefined field name "whatzit"
    # output msg: Found undefined field name 'whatzit' when expecting ['name', 'description', 'sections', 'parent_specs', 'child_specs', 'requirements'] as defined in RequirementSpecification
    def test_undefined_field_name(self):
        """Test check_aac_impl.py with an undefined field."""
        with tempfile.TemporaryDirectory() as temp_dir:
            aac_file_path = os.path.join(os.path.dirname(__file__), "undefined_field_name.aac")
            temp_aac_file_path = os.path.join(temp_dir, "my_plugin.aac")
            shutil.copy(aac_file_path, temp_aac_file_path)

            check_args = [temp_aac_file_path]

            exit_code, output_message = self.run_check_cli_command_with_args(check_args)
            # Assert for return code 6 (undefined field name)
            self.assertEqual(6, exit_code, f"Expected to fail but ran successfully with message: {output_message}")
            self.assertNotIn("My plugin was successful.", output_message)  # only appears when --verbose is passed in.
            self.assertIn("my_plugin.aac", output_message)



"""AaC Plugin implementation module for the Version plugin."""

from os import path, mkdir
from typing import Callable
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
)

plugin_name = "Gen Plugin"


def before_gen_plugin_check(
    aac_plugin_file: str,
    code_output: str,
    test_output: str,
    doc_output: str,
    no_prompt: bool,
    force_overwrite: bool,
    evaluate: bool,
    run_check,
) -> ExecutionResult:
    """Run the Check AaC command before the gen-plugin command."""

    return run_check(aac_plugin_file, False, False)


def gen_plugin(
    aac_plugin_file: str,
    code_output: str,
    test_output: str,
    doc_output: str,
    no_prompt: bool,
    force_overwrite: bool,
    evaluate: bool,
) -> ExecutionResult:
    """Gen-plugin command entry point."""

    # There really isn't any specific behavior here, but we need to return a result
    return ExecutionResult(plugin_name, "gen-plugin", ExecutionStatus.SUCCESS, [])


def after_gen_plugin_generate(
    aac_plugin_file: str,
    code_output: str,
    test_output: str,
    doc_output: str,
    no_prompt: bool,
    force_overwrite: bool,
    evaluate: bool,
    generate: Callable,
) -> ExecutionResult:
    """Generate the plugin by running generate command."""
    gen_plugin_generator_file = path.abspath(
        path.join(path.dirname(__file__), "./gen_plugin_generator.aac")
    )
    return generate(
        aac_plugin_file,
        gen_plugin_generator_file,
        code_output,
        test_output,
        doc_output,
        no_prompt,
        force_overwrite,
        evaluate,
    )


def before_gen_project_check(
    aac_project_file: str,
    output: str,
    no_prompt: bool,
    force_overwrite: bool,
    evaluate: bool,
    run_check,
) -> ExecutionResult:
    """Run the Check AaC command before the gen-project command."""

    return run_check(aac_project_file, False, False)


def gen_project(
    aac_project_file: str,
    output: str,
    no_prompt: bool,
    force_overwrite: bool,
    evaluate: bool,
) -> ExecutionResult:
    """Gen-project command entry."""

    # There really isn't any specific behavior here, but we need to return a result
    return ExecutionResult(plugin_name, "gen-project", ExecutionStatus.SUCCESS, [])


def after_gen_project_generate(
    aac_project_file: str,
    output: str,
    no_prompt: bool,
    force_overwrite: bool,
    evaluate: bool,
    generate: Callable,
) -> ExecutionResult:
    """Generate the project by running generate command."""
    gen_plugin_generator_file = path.abspath(
        path.join(path.dirname(__file__), "./gen_plugin_generator.aac")
    )
    result = generate(
        aac_project_file,
        gen_plugin_generator_file,
        output,
        output,
        output,
        no_prompt,
        force_overwrite,
        evaluate,
    )
    if result.is_success():
        src_path = path.join(output, "src")
        tests_path = path.join(output, "tests")
        docs_path = path.join(output, "docs")
        if not path.exists(src_path):
            mkdir(src_path)
        if not path.exists(tests_path):
            mkdir(tests_path)
        if not path.exists(docs_path):
            mkdir(docs_path)

    return result
